# Efficient Hierarchical Decomposition of Repetitive Traces for ML-Driven Analysis
This repository implements an algorithm for finding the longest possible repetitions in a given character array using rolling hashes. The functionality of this algorithm is explained in detail in the associated paper, 'Efficient Hierarchical Decomposition of Repetitive Traces for ML-Driven Analysis'. Optionally, the algorithm can be parallelized, and the paper discusses the benefits and limitations of this feature.

## Dependencies
The following dependencies are required to build and run the software:

- [Python](https://www.python.org/) 3.9.12 or higher
- [Meson](https://mesonbuild.com/) 0.63.3 or higher
- [OpenMP](https://www.openmp.org/) 4.5 or higher (Only required for parallelization)

## Tool Interfaces

The tool exposes two primary interfaces, which are implemented in the following Python files:

- [`cimpl.py`](python/cimpl.py)
- [`cimpl_omp.py`](python/cimpl_omp.py)

These are described in more detail in the [Using the Software](#using-the-software) section below. Additionally, you can use the `detect_repetions()` function directly from [`repetition.c`](src/repetition.c) by including the header file [`repetition.h`](include/repetition.h). The function requires the following arguments:

- `repetitionNode *result`: A linked list to store the result.
- `char signs[]`: An array of symbols to be analyzed.
- `int length`: The length of the `signs[]` array.

By default, the algorithm runs without parallelization. To enable parallelization, define `#define _OPENMP`. The data structure `repetitionNode` and the `printRepNode()` function for displaying results are also available by including [`repetition.h`](include/repetition.h).

## Building

The tool uses Meson to simplify the build process. To build the tool, follow these steps:

```sh
meson setup build                           # Set up the build environment and configure the project in the 'build' directory
cd build                                    # Change to the 'build' directory
ninja                                       # Build the tool
cd ..                                       # Return to the project root
export LD_LIBRARY_PATH=$(readlink -f build) # Specify the library location
export PYTHONPATH=$(readlink -f python)     # Specify the Python modules location
```

After running these commands, the tool should be ready to use. Make sure that all dependencies are installed and meet the required versions.


## Using the Software
You can use the tool either with or without parallelization. The following sections explain how to use the tool in both modes.

### Without Parallelization
To use the tool without parallelization, call the function `detect_repetitions_cimpl(data)` from [`cimpl.py`](python/cimpl.py). The input should be a byte array. This function internally calls the `detect_repetitions()` function from [`repetition.c`](src/repetition.c). The results are returned as a linked list in `rn`, which can be printed using the `print_rep_node()` function.

Example:

```py
from cimpl import detect_repetitions_cimpl, print_rep_node

data = b'ababc'

result = detect_repetitions_cimpl(data)         # Finds repetitions in data

print_rep_node(result)                          # prints the found repetitions                
```
Output:
```sh
((ab)^2 c)^1
```

### With Parallelization
To use the tool with parallelization, import the module [`cimpl_omp.py`](python/cimpl_omp.py). The usage is very similar to the non-parallelized version, but you can specify the number of threads for parallelization using the function `num_threads_for_C(amount)`. This function sets the number of cores to be used in the parallelization process.

Example:
```py
from cimpl_omp import detect_repetitions_cimpl, print_rep_node, num_threads_for_C

data = b'ababc'

num_threads_for_C(32)                           # Sets the number of threads (cores) for parallelization to 32
result = detect_repetitions_cimpl(data)         # Finds repetitions in data
print_rep_node(result)                          # Prints the found repetitions
```
Output:
```sh
((ab)^2 c)^1
```

Note: The parallelization flag `_OPENMP` is set internally by `num_threads_for_C`.

### Using the C-Interface directly
To use the C-Interface directly, you need to include [`repetition.h`](include/repetition.h) as described earlier. After including the header file, create a linked list to hold the results by initializing a `repetitionNode` using the `initRepetitionNode()` function. Then, call the `detect_repetitions()` function to find repetions in the data. This function takes the parameters described [previously](#tool-interfaces). Finally, use `printRepNode()` to print the results. 
Example:
```c
#include <stdio.h>
#include <repetition.h>

int main(){
    char data[] = {'a', 'b', 'a', 'b', 'c'};
    int length = sizeof(data)/sizeof(data[0]);

    repetitionNode* result = initRepetitionNode(1); // Initialize linked list for the results

    detect_repetitions(result, data, length);       // Finds repetitions in data

    printRepNode(result);                           // Prints the found repetitions
    return 0;
}
```
Output:
```sh
((ab)^2 c)^1
```

## Parallelization
The algorithm uses a rolling hash to compute hash values for consecutive bytes and stores these values in a separate array. A repetition is detected when two consecutive hashes are identical. If no repetitions are found, the algorithm calculates new hashes by removing the last byte from each consecutive sequence. This process leverages the rolling hash property, where the hash value can be updated efficiently from the previous value.

Parallelization is applied during the initial hashing and when removing the last byte. The number of threads can be controlled using the `num_threads_for_C(amount)` function, which defines how many cores will be used. Keep in mind that using more cores does not always equate to better performance. While parallelization can speed up the algorithm, too many threads may actually decrease performance due to the overhead of managing parallel tasks. For further details, see the referenced paper.