import random
import timeit
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
import multiprocessing

def generate_random_bytes(length, seed=42, minimum=0x41, maximum=0x59):
    random.seed(seed)
    return bytes([random.randint(minimum, maximum) for _ in range(length)])

def generate_random_hier(length, seed, depth, maxlen_seq, minimum=0x41, maximum=0x59, flat_prob=0.1):
    random.seed(seed)

    if depth == 0:
        return generate_random_bytes(length, random.random(), minimum, maximum)

    remaining = length
    result = b''

    while remaining > 0:
        choice = random.uniform(0, 1)
        seq_len = random.randint(0, min(remaining, maxlen_seq))
        if choice < flat_prob or seq_len // 2 <= 0:
            # flat
            result += generate_random_bytes(seq_len, random.random(), minimum, maximum)
        else:
            # hier
            hier_len = random.randint(1, seq_len//2)
            num = seq_len // hier_len
            appended_str = generate_random_hier(hier_len, random.random(), depth - 1, maxlen_seq, minimum, maximum, flat_prob) * num 
            result += appended_str
            if len(appended_str) <= 0:
                assert(False)
        remaining = length - len(result)
        if remaining < 0:
            assert(False)
    return result

def performance_test(params):
    if params[0]:
        length, seed, depth, maxlen_seq = params[1:]
        data = generate_random_hier(length, seed, depth, maxlen_seq)
    else:
        length, seed  = params[1:]
        data = generate_random_bytes(length, seed)

    if len(data) != length:
        assert(False)


    setup_code = f'''
from repetition import detect_repetitions
data = {data}
'''
    test_code = 'x = detect_repetitions(data)'
    duration = timeit.timeit(test_code, setup=setup_code, number=1)

    setup_code_cimpl = f'''
from cimpl import detect_repetitions_cimpl
data = {data}
'''
    test_code_cimpl = 'x = detect_repetitions_cimpl(data)'
    duration_cimpl = timeit.timeit(test_code_cimpl, setup=setup_code_cimpl, number=1)

    if params[0]:
        print(f"Length: {length}, Seed: {seed}, Depth: {depth}, Maxlen_seq: {maxlen_seq}, Duration: {duration} seconds, Duration in C: {duration_cimpl} seconds")
    else:
        print(f"Length: {length}, Seed: {seed}, Duration: {duration} seconds, Duration in C: {duration_cimpl} seconds")

    if params[0]:
        return length, depth, maxlen_seq, duration, duration_cimpl, data
    else:
        return length, duration, duration_cimpl, data

# def fit_quadratic_curve(data):
#     def quadratic(x, a, b, c):
#         return a * x**2 + b * x + c
#     lengths = data['length'].values
#     durations = data['duration'].values
#     popt, _ = curve_fit(quadratic, lengths, durations)
#     return popt

def plot_results(data_hier, data_random):
    sns.set_theme(style='ticks')

    plt.figure(dpi=200)

    data_hier['kind'] = 'Hierachical'
    data_random['kind'] = 'Random'

    data_combined = pd.concat((data_hier, data_random))

    l_duration = 'Runtime [s]'
    l_kind = 'Sample Type'
    l_length = 'Total Length of Sample Sequence'

    data_combined = data_combined.rename(columns={
        'kind': l_kind,
        'duration': l_duration,
        'length': l_length,
    })

    print(data_combined.columns)

    g = sns.FacetGrid(data_combined, col=l_kind, height=4, aspect=1)
    g.map_dataframe(sns.boxplot, x=l_length, y=l_duration)

    sns.despine()

    plt.savefig('performance-combined.png')

def plot_results_C(data_hier, data_random):
    # Made a new function because I wasn't sure whether panda could plot diffrent data into on Plot
    # can also be reversed and put in one function if it is possible
    sns.set_theme(style='ticks')

    plt.figure(dpi=200)

    data_hier['kind'] = 'Hierachical'
    data_random['kind'] = 'Random'

    data_combined = pd.concat((data_hier, data_random))

    l_duration_C = 'Runtime [s]'
    l_kind = 'Sample Type'
    l_length = 'Total Length of Sample Sequence'

    data_combined = data_combined.rename(columns={
        'kind': l_kind,
        'duration_C': l_duration_C,
        'length': l_length,
    })

    print(data_combined.columns)

    g = sns.FacetGrid(data_combined, col=l_kind, height=4, aspect=1)
    g.map_dataframe(sns.boxplot, x=l_length, y=l_duration_C)

    sns.despine()

    plt.savefig('performance-combined-C.png')


pool = multiprocessing.Pool(multiprocessing.cpu_count())

def main(hier=True, pool_a=True):
    # Define parameters for sweeping
    lengths = np.linspace(0, 5000, 5)  # Adjust as needed
    seeds = list(range(16))
    depths = [4, 8, 16, 32]
    maxlen_seqs = [.25, .5, .75, 1]

    if hier:
        params = [(True, int(length), int(seed * length - depth), depth, int(maxlen_seq * length))
                for length in lengths
                for seed in seeds
                for depth in depths
                for maxlen_seq in maxlen_seqs]
        if pool_a:
            results = pool.map(performance_test, params)
        else:
            results = map(performance_test, params)
        data = pd.DataFrame(results, columns=['length', 'depth', 'maxlen_seq', 'duration', 'duration_C', 'data'])

        print("finished hier")

        data.to_pickle('results-hier.pkl')

    else:
        seeds = range(16 * 4 * 4)
        params = [(False, int(length), int(seed * length))
                for length in lengths
                for seed in seeds]
        if(pool_a):
            results = pool.map(performance_test, params)
        else:
            results = map(performance_test, params)
        data = pd.DataFrame(results, columns=['length', 'duration', 'duration_C', 'data'])
        data.to_pickle('results-random.pkl')

        print("finished random")

    # plot_results(data)

import argparse

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--run', action='store_true')
    parser.add_argument('--hier', action='store_true')
    parser.add_argument('--random', action='store_true')
    parser.add_argument('--plot', action='store_true')
    parser.add_argument('--pool_a', action='store_true')

    args = parser.parse_args()

    if args.hier or args.run:
        if(args.pool_a):
            main(hier=True, pool_a=True)
        else:
            main(hier=True, pool_a=False)

    if args.random or args.run:
        if(args.pool_a):
            main(hier=False, pool_a=True)
        else:
            main(hier=False, pool_a=False)

    if args.plot:
        plot_results(pd.read_pickle('results-hier.pkl'), pd.read_pickle('results-random.pkl'))
        plot_results_C(pd.read_pickle('results-hier.pkl'), pd.read_pickle('results-random.pkl'))

